from pathlib import Path
import pytest
from ewoksorange.tests.utils import execute_task

from ewoks3dxrd.tasks.segment_frame import SegmentFrame


def test_segment_frame(inp_config):

    # if you change this config setting,
    # assertion statement assert len(found_peaks[0]) == 76 will fail

    segmenter_config = {
        "threshold": 70,
        "smooth_sigma": 1.0,
        "bgc": 0.9,
        "min_px": 3,
        "offset_threshold": 100,
        "ratio_threshold": 150,
    }

    file_cor_config = {
        "bg_file": inp_config.get("bg_file"),
        "mask_file": inp_config.get("mask_file"),
        "flat_file": None,
        "dark_file": None,
    }

    scan_folder = Path(inp_config.get("scan_folder"))
    sample_folder_config = {
        "detector": "frelon3",
        "omega_motor": "diffrz",
        "scan_folder": str(scan_folder),
        "analyse_folder": inp_config.get("analyse_folder"),
    }

    inputs = {
        "folder_config": sample_folder_config,
        "segmenter_algo_params": segmenter_config,
        "correction_files": file_cor_config,
    }
    task = SegmentFrame(inputs=inputs)
    task.execute()
    outputs = task.outputs

    raw_image = outputs["raw_image"]
    bg_corrected_image = outputs["bg_corrected_image"]
    found_peaks = outputs["found_peaks"]
    assert raw_image is not None
    assert bg_corrected_image is not None
    assert found_peaks is not None
    assert len(found_peaks) == 2
    assert len(found_peaks[0]) == 76


def test_wrong_threshold(inp_config):
    segmenter_config = {
        "threshold": "not_a_valid_threshold",
        "smooth_sigma": 1.0,
        "bgc": 0.9,
        "min_px": 3,
        "offset_threshold": 100,
        "ratio_threshold": 150,
    }
    file_cor_config = {
        "bg_file": inp_config.get("bg_file"),
        "mask_file": inp_config.get("mask_file"),
        "flat_file": None,
        "dark_file": None,
    }

    scan_folder = Path(inp_config.get("scan_folder"))
    sample_folder_config = {
        "detector": "frelon3",
        "omega_motor": "diffrz",
        "scan_folder": str(scan_folder),
        "analyse_folder": inp_config.get("analyse_folder"),
    }

    with pytest.raises(RuntimeError) as exc_info:
        execute_task(
            SegmentFrame,
            inputs={
                "folder_config": sample_folder_config,
                "segmenter_algo_params": segmenter_config,
                "correction_files": file_cor_config,
            },
        )
    assert "Input should be a valid integer" in str(exc_info.value.__cause__)



from ewokscore import Task
from pathlib import Path
from ImageD11 import frelon_peaksearch

from ..models import (
    SegmenterConfig,
    SegmenterFolderConfig,
    SegmenterCorrectionFiles,
)
from ..utils import extract_sample_info, get_frame_image


class SegmentFrame(
    Task,
    input_names=["folder_config", "segmenter_algo_params", "correction_files"],
    output_names=[
        "raw_image",
        "bg_corrected_image",
        "found_peaks",
    ],
):
    """
    This task produces single segmented frame
    """

    def run(self):

        seg_folder_config = SegmenterFolderConfig(**self.inputs.folder_config)
        detector = seg_folder_config.detector
        scan_folder = seg_folder_config.scan_folder
        _, sample_name, dset_name, scan_number = extract_sample_info(
            path_str=scan_folder
        )
        masterfile_path = Path(scan_folder).parent / f"{sample_name}_{dset_name}.h5"
        raw_image = get_frame_image(
            file_path=masterfile_path,
            detector=detector,
            scan_id=scan_number + ".1",
        )
        raw_image = raw_image.astype("uint16")

        segmenter_cfg = SegmenterConfig(**self.inputs.segmenter_algo_params)
        correction_files_config = SegmenterCorrectionFiles(
            **self.inputs.correction_files
        )
        segmenter_settings = {
            "bgfile": correction_files_config.bg_file,
            "maskfile": correction_files_config.mask_file,
            "darkfile": correction_files_config.dark_file,
            "flatfile": correction_files_config.flat_file,
            "threshold": segmenter_cfg.threshold,
            "smoothsigma": segmenter_cfg.smooth_sigma,
            "bgc": segmenter_cfg.bgc,
            "minpx": segmenter_cfg.min_px,
            "m_offset_thresh": segmenter_cfg.offset_threshold,
            "m_ratio_thresh": segmenter_cfg.ratio_threshold,
        }
        image_worker = frelon_peaksearch.worker(**segmenter_settings)
        goodpeaks = image_worker.peaksearch(img=raw_image, omega=0)
        # 23 and 24 are the columns for fast_column index and slow column index
        peak_positions = goodpeaks[:, 23:25].T

        self.outputs.raw_image = raw_image
        self.outputs.bg_corrected_image = image_worker.smoothed
        self.outputs.found_peaks = peak_positions



from ewokscore import Task
from pathlib import Path
import shutil

from ImageD11 import columnfile as PeakColumnFile


class GeometryTransformation(
    Task,
    input_names=[
        "spatial_corrected_peaks_3d_file",
        "geometry_par_file",
    ],
    output_names=["geometry_updated_3d_peaks_file"],
):
    """
    Does the geometry Transformation on the detector spatial corrected 3d peaks file
    and geometry parameter (.par) file,
    Inputs:
        '3d_peaks_file'
        a detector spatial corrected 3d peaks columnfile as .h5 and
        geometry file path a .par file
    This task performs the following operations:
    1. Gathers geometry information from the `geometry_tdxrd.par` file.
    2. Copy the geometry file in the directory structure:
            `analysis_folder / dset_name / (dset_name + sample_name)`.
            i.e the parent folder of 'detector_spatial_corrected_3d_peaks_file'
    3. Applies the geometry correction to the 3D peaks column file using
        the `ImageD11` dataset class.
    4. Save the geometry corrected 3d peaks columnfile (output: `geometry_updated_3d_peaks_file`)
    """

    def run(self):
        seg_3d_col_file = self.inputs.spatial_corrected_peaks_3d_file
        geometry_par_file = self.inputs.geometry_par_file

        error_message = []
        if not Path(geometry_par_file).exists():
            error_message.append(f"Geometry file '{geometry_par_file}' not found.")

        if not geometry_par_file.endswith(".par"):
            error_message.append(
                f"Invalid geometry file '{geometry_par_file}'. Expected a '.par' file."
            )

        if not Path(seg_3d_col_file).exists():
            error_message.append(
                f"Provided '{seg_3d_col_file}' spatial corrected and segmented 3D peak file is missing."
            )

        if error_message:
            raise ValueError("\n".join(error_message))

        work_folder = Path(seg_3d_col_file).parent
        par_folder = work_folder / "par_folder"
        par_folder.mkdir(exist_ok=True)
        new_geometry_par_file = par_folder / "geometry_tdxrd.par"
        shutil.copy2(geometry_par_file, new_geometry_par_file)

        # following code is derived by refering:
        # https://github.com/FABLE-3DXRD/ImageD11/blob/master/ImageD11/sinograms/dataset.py
        # refer at function  update_colfile_pars()
        # https://github.com/FABLE-3DXRD/ImageD11/blob/master/ImageD11/parameters.py
        # refer at function loadparameters() in the class 'parameters' (yes it is a class not function)
        # https://github.com/FABLE-3DXRD/ImageD11/blob/master/ImageD11/columnfile.py
        # refer at function updateGeometry() in the class 'columnfile' (yes it is a class not function)
        # as for as my investigation, it does not do any computation,
        # but with the style of imageD11 library, it adds various geometry parameter from
        # geometry_tdxrd.par file into the column file,
        columnfile_3d = PeakColumnFile.columnfile(filename=seg_3d_col_file)
        columnfile_3d.parameters.loadparameters(filename=str(new_geometry_par_file))
        columnfile_3d.updateGeometry()
        geometry_updated_3d_peaks_file = work_folder / "geometry_updated_3d_peaks.h5"
        PeakColumnFile.colfile_to_hdf(
            columnfile_3d, str(geometry_updated_3d_peaks_file)
        )
        self.outputs.geometry_updated_3d_peaks_file = str(
            geometry_updated_3d_peaks_file
        )



from ewokscore import Task
from pathlib import Path

from ImageD11 import columnfile as PeakColumnFile
from ImageD11.peakselect import sorted_peak_intensity_mask_and_cumsum


class FilterByIntensity(
    Task,
    input_names=[
        "peaks_3d_file",
        "intensity_frac",
    ],
    output_names=[
        "Ifilt_h5_col_file",
        "Ifilt_flt_col_file",
    ],
):
    """
    Does the Intensity based peaks filter,
        computes intensity metric based on sum_intensity, ds (reciprocal distance) columns from the input file
        normalize with the maximum value of intensity metric,
        only keeps the rows whose value is above the given input 'intensity_frac'
        and save them in ascii (.flt) and .h5 format.

    Inputs:

    - `peaks_3d_file`: a 3d merged column file h5 file, it is must be corrected for geometry and detector
    - `intensity_frac`: float value to remove the peaks row whose intensity metric below than this value.

    Outputs:

    - `Ifilt_h5_col_file`:  the rest of the strong peaks were saved as .h5 file path.
    - `Ifilt_flt_col_file`: the same .h5 file in .flt (ascii) format.
    """

    def run(self):
        peaks_3d_file = Path(self.inputs.peaks_3d_file)
        intensity_frac = self.inputs.intensity_frac
        if not peaks_3d_file.exists():
            raise FileNotFoundError(
                f"Input 3d peaks file {peaks_3d_file} does not exist."
            )

        cf = PeakColumnFile.columnfile(filename=str(peaks_3d_file))
        mask, _ = sorted_peak_intensity_mask_and_cumsum(colf=cf, frac=intensity_frac)
        cf.filter(mask)
        file_tag = f"{intensity_frac:.4f}".replace(".", "p")
        flt_file_path = (
            peaks_3d_file.parent / f"intensity_frac_{file_tag}_{peaks_3d_file.stem}.flt"
        )
        cf.writefile(str(flt_file_path))
        h5_file_path = flt_file_path.with_suffix(".h5")
        PeakColumnFile.colfile_to_hdf(cf, str(h5_file_path))
        self.outputs.Ifilt_h5_col_file = str(h5_file_path)
        self.outputs.Ifilt_flt_col_file = str(flt_file_path)



import pytest

from ewoks3dxrd.utils import read_lattice_cell_data


@pytest.mark.parametrize("space_group", (229, "I"))
def test_read_lattice_cell_data(tmp_path, space_group):
    par_file = tmp_path / "Fe.par"
    with open(par_file, "w") as f:
        f.write(
            f"""cell__a 2.8694
cell__b 2.8694
cell__c 2.8694
cell_alpha 90.0
cell_beta 90.0
cell_gamma 90.0
cell_lattice_[P,A,B,C,I,F,R] {space_group}
"""
        )

    unit_cell_parameters = read_lattice_cell_data(par_file)

    assert unit_cell_parameters.a == 2.8694
    assert unit_cell_parameters.b == 2.8694
    assert unit_cell_parameters.c == 2.8694
    assert unit_cell_parameters.alpha == 90
    assert unit_cell_parameters.beta == 90
    assert unit_cell_parameters.gamma == 90
    assert unit_cell_parameters.space_group == space_group



import os
from pathlib import Path

import h5py
import pytest
from ewoksorange.canvas.handler import OrangeCanvasHandler
from ewoksorange.tests.conftest import qtapp  # noqa F401
from ImageD11.grain import read_grain_file


@pytest.fixture(scope="session")
def ewoks_orange_canvas(qtapp):  # noqa F811
    with OrangeCanvasHandler() as handler:
        yield handler


@pytest.fixture
def inp_config(tmp_path):
    scan_folder = (
        "/data/projects/id03_3dxrd/expt/RAW_DATA/FeAu_0p5_tR/FeAu_0p5_tR_ff1/scan0001"
    )

    if not os.path.exists(scan_folder):
        raise FileNotFoundError(
            f"""
            Could not find {scan_folder}.
            Before running this test, be sure to have a link to /data/projects at /data/:
             - mkdir -p /data/projects/id03_3dxrd
             - ln -s /gpfs/.../data/projects /data
            Sudo rights might be needed.
            """
        )

    return {
        "detector": "frelon3",
        "omega_motor": "diffrz",
        "dty_motor": "diffty",
        "bg_file": None,
        "mask_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/mask.edf",
        "spline_file": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/frelon36.spline",
        "e2dx_file": None,
        "e2dy_file": None,
        "scan_folder": scan_folder,
        "analyse_folder": os.path.join(tmp_path, "test_my_task"),
        "stateful_imageD11_file": None,
    }


def assert_indexing_results(indexing_task_outputs):
    ubi_file_path = indexing_task_outputs["grain_ubi_file"]
    assert Path(ubi_file_path).exists()
    h5_file_path = indexing_task_outputs["grain_nexus_file"]
    assert Path(h5_file_path).exists()
    with h5py.File(h5_file_path, "r") as f:
        assert f["entry/indexed_grains/grains/UBI"][()].shape == (59, 3, 3)
        assert f["entry/indexed_grains/grains/translation"][()].shape == (59, 3)


def assert_grain_map_results(grain_map_task_outputs):
    ascii_grains_file = grain_map_task_outputs["ascii_grain_map_file"]
    assert Path(ascii_grains_file).exists()
    list_grains = read_grain_file(ascii_grains_file)
    assert len(list_grains) == 52



from ewoks3dxrd.tasks.index_grains import IndexGrains

from .conftest import assert_indexing_results


def test_indexing():
    filepath = "/data/projects/id03_3dxrd/ewoks_test_data/indexing/intensity_frac_0p9837_Fe_1p0100_filtered_3d_peaks.h5"

    inputs = {
        "flt_peaks_3d_file": filepath,
        "reciprocal_dist_tol": 0.05,
        "gen_rings_from_idx": (0, 1),
        "score_rings_from_idx": (0, 1, 2, 3),
        "lattice_name": "Fe",
        "hkl_tols": (0.01, 0.02, 0.03, 0.04),
        "min_pks_frac": (0.9, 0.75),
    }

    task = IndexGrains(inputs=inputs)
    task.execute()

    assert_indexing_results(task.outputs)



import sys
from pathlib import Path
from typing import Literal, Optional, Tuple, Union

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from pydantic import BaseModel, Field, field_validator


class SampleConfig(BaseModel):
    omega_motor: Literal["diffrz", "omega"]
    scan_folder: str

    @field_validator("scan_folder")
    @classmethod
    def path_exists(cls, path: Optional[str]):
        if path is not None and not Path(path).exists():
            raise ValueError(f"Supplied path {path} does not exist.")
        return path


class SegmenterFolderConfig(SampleConfig):
    detector: Literal["frelon1", "frelon3"]
    analyse_folder: str


class SegmenterConfig(BaseModel):
    threshold: int
    smooth_sigma: float
    bgc: float
    min_px: int
    offset_threshold: int
    ratio_threshold: int


class SegmenterCorrectionFiles(BaseModel):
    bg_file: Optional[str] = None
    mask_file: Optional[str] = None
    dark_file: Optional[str] = None
    flat_file: Optional[str] = None

    @field_validator("bg_file", "mask_file", "dark_file", "flat_file")
    @classmethod
    def path_exists(cls, path: Optional[str]):
        if path is not None and not Path(path).exists():
            raise ValueError(f"Supplied path {path} does not exist.")
        return path


class UnitCellParameters(BaseModel):
    a: float = Field(alias="cell__a")
    b: float = Field(alias="cell__b")
    c: float = Field(alias="cell__c")
    alpha: float = Field(alias="cell_alpha")
    beta: float = Field(alias="cell_beta")
    gamma: float = Field(alias="cell_gamma")
    space_group: Union[
        Literal["P", "A", "B", "C", "I", "F", "R"], Annotated[int, Field(ge=1, le=230)]
    ] = Field(alias="cell_lattice_[P,A,B,C,I,F,R]")

    @property
    def lattice_parameters(self) -> Tuple[float, float, float, float, float, float]:
        return self.a, self.b, self.c, self.alpha, self.beta, self.gamma



from __future__ import annotations

import os
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import List, Tuple, Generator

import h5py
import numpy as np
from ImageD11 import blobcorrector
from ImageD11 import columnfile as PeakColumnFile
from ImageD11.grain import grain as Grain
from ImageD11.refinegrains import refinegrains as RefineGrains
from tqdm import tqdm

from .models import UnitCellParameters


class TqdmProgressCallback(tqdm):
    def __init__(self, *args, TaskInstance=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.TaskInstance = TaskInstance
        self.finished_count = 0

    def update(self, n: int = 1):
        super().update(n)
        self.finished_count += n
        if self.TaskInstance:
            self.TaskInstance.progress = 100.0 * (self.finished_count / self.total)


def correct_column_file(filename: str | Path, correction_files: Tuple[str, str] | str):
    segmented_columnfile_3d = PeakColumnFile.columnfile(filename=filename)
    peak_3d_dict = {
        title: segmented_columnfile_3d[title]
        for title in segmented_columnfile_3d.keys()
    }
    peak_3d_dict["spot3d_id"] = np.arange(len(peak_3d_dict["s_raw"]))
    raw_columnfile_3d = PeakColumnFile.colfile_from_dict(peak_3d_dict)

    if isinstance(correction_files, str):
        splinefile = correction_files
        return blobcorrector.correct_cf_with_spline(raw_columnfile_3d, splinefile)

    if isinstance(correction_files, tuple) and len(correction_files) == 2:
        e2dxfile, e2dyfile = correction_files
        return blobcorrector.correct_cf_with_dxdyfiles(
            raw_columnfile_3d, e2dxfile, e2dyfile
        )

    raise ValueError(
        f"Detector Spatial correction cannot be performed unless a spline file or a couple of e2dx, ed2dy files is provided. Got {correction_files}."
    )


def load_par_file(filepath: str | Path) -> dict:
    """
    Read .par file, and return each line as key, value pair
    """
    read_parameters = {}

    with open(filepath, "r") as file:
        for line in file:
            parts = line.strip().split()
            if len(parts) != 2:
                raise ValueError(f"Invalid line format: {line.strip()}")
            key, value = parts
            read_parameters[key] = value

    return read_parameters


def read_lattice_cell_data(lattice_data_file_path: str | Path) -> UnitCellParameters:
    """
    Reads lattice cell data from a file and extracts the lattice parameters and space group.
    """
    return UnitCellParameters(**load_par_file(lattice_data_file_path))


def save_grains_to_nexus(
    nexus_filename: str | Path,
    geometry_par_path: str | Path,
    lattice_par_file: str | Path,
    lattice_name: str,
    grains: List[Grain],
) -> None:
    """
    Save grain information (UBI matrices, lattice parameters) and
    the experiment geometry into a NeXus file.

    Parameters:
    - nexus_filename: Output NeXus file path
    - geometry_par_path: Path to the geometry_tdxrd.par file
    - lattice_par_file: Path to lattice parameters file
    - lattice_name: Name of the lattice type
    - grains (list of grain instances): List of grain objects with:
        - grain.ubi: (3,3) array
    """

    unit_cell_parameters = read_lattice_cell_data(
        lattice_data_file_path=lattice_par_file
    )
    lattice_parameters, space_group = (
        unit_cell_parameters.lattice_parameters,
        unit_cell_parameters.space_group,
    )

    with open(geometry_par_path, "r") as f:
        geometry_lines = [line.strip() for line in f]
    geometry_par = "\n".join(geometry_lines)

    num_grains = len(grains)
    ubi_matrices = np.zeros((num_grains, 3, 3), dtype=grains[0].ubi.dtype)
    translations = np.full((num_grains, 3), fill_value=np.nan)

    for i, grain in enumerate(grains):
        ubi_matrices[i] = grain.ubi
        if grain.translation:
            translations[i] = grain.translation

    with h5py.File(nexus_filename, "w") as f:
        entry = f.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        index_grains = entry.create_group("indexed_grains")
        index_grains.attrs["NX_class"] = "NXprocess"
        parameters_group = index_grains.create_group("parameters")
        parameters_group.attrs["NX_class"] = "NXcollection"
        parameters_group.create_dataset("geometry_par", data=geometry_par)
        parameters_group.create_dataset("lattice_name", data=lattice_name)
        parameters_group.create_dataset(
            "lattice_parameters", data=np.array(lattice_parameters)
        )
        if isinstance(space_group, int):
            parameters_group.create_dataset("space_group_number", data=space_group)
        else:
            parameters_group.create_dataset("space_group_symbol", data=space_group)

        grains_group = index_grains.create_group("grains")
        grains_group.attrs["NX_class"] = "NXdata"
        grains_group.create_dataset("UBI", data=ubi_matrices)
        grains_group.create_dataset("translation", data=translations)


def read_wavelength(geometry_par_file: str | Path) -> float:
    with open(geometry_par_file, "r") as f:
        for line in f:
            if line.startswith("wavelength"):
                try:
                    wavelength = float(line.strip().split()[1])
                    break
                except (IndexError, ValueError):
                    raise ValueError(
                        f"Could not found a valid wavelength field in {geometry_par_file}"
                    )
    return wavelength


def get_omega_slope(filepath, scan_number, omega_motor):
    with h5py.File(filepath, "r") as hin:
        omega_angles = hin[f"{scan_number}.1/measurement"].get(omega_motor, None)
        omega_array = omega_angles[()]

    omegas_sorted = np.sort(omega_array)
    omega_step = np.round(np.diff(omegas_sorted).mean(), 3)
    return omega_step / 2


def refine_grains(
    tolerance: float,
    intensity_tth_range: Tuple[float, float],
    omega_slope: float,
    symmetry: str,
    parameter_file: str | Path,
    filtered_peaks_file: str | Path,
    ubi_file: str | Path,
) -> RefineGrains:
    refined_grains = RefineGrains(
        tolerance=tolerance,
        intensity_tth_range=intensity_tth_range,
        OmSlop=omega_slope,
    )
    refined_grains.loadparameters(str(parameter_file))
    refined_grains.loadfiltered(str(filtered_peaks_file))
    refined_grains.readubis(str(ubi_file))
    refined_grains.makeuniq(symmetry)
    refined_grains.generate_grains()
    refined_grains.refinepositions()

    return refined_grains


def extract_sample_info(path_str: str) -> Tuple[str, str, str, str]:
    """
    Helper function to extract the dataroot, sample, dset_name, scan no from the
    scan folder path
    """
    path = Path(path_str)
    if len(path.parts) < 4:
        raise ValueError(
            "Expected path structure to be of the for `/dataroot/sample_name/dset_name/scan_number`"
        )

    # check provided path is actually a scan folder
    if len(path_str) < 8 or path_str[-8:-4] != "scan":
        raise ValueError(
            f"Invalid scan folder path: '{path_str}' should end with `scanxxxx` where xxxx are numbers"
        )

    # Extract scan number (last 4 digits)
    try:
        scan_number = int(path.name[-4:])
    except (ValueError, IndexError):
        raise ValueError(f"Expected {path.name} to end with 4 numbers.")

    # Extract dset name
    dset_folder_name = path.parent.name
    if "_" not in dset_folder_name:
        raise ValueError(
            f"Invalid dataset name: no underscore found in {dset_folder_name} to separate sample and dataset name.",
        )

    dataroot = path.parent.parent.parent.__str__()
    sample_name = path.parent.parent.name
    dset_name = dset_folder_name[len(sample_name + "_") :]
    return dataroot, sample_name, dset_name, str(scan_number)


def get_frame_image(
    file_path: str | Path,
    detector: str,
    scan_id: str,
    frame_idx: int | None = None,
    counter: str = "_roi1",
):
    """
    extract a frame image from master file path
    """
    with h5py.File(file_path, "r") as h5file:
        if frame_idx is None:
            detector_ctr = detector + counter
            frame_idx = np.argmax(h5file[f"{scan_id}/measurement/{detector_ctr}"][:])
        return h5file[f"{scan_id}/measurement/{detector}"][frame_idx]


@contextmanager
def tmp_processing_files(
    initial_ubi_file: str | Path, geo_par_file: str | Path, lattice_par_file: str | Path
) -> Generator[Tuple[str, str], None, None]:
    _, tmp_ubi_file = tempfile.mkstemp()
    _, tmp_par_file = tempfile.mkstemp()

    try:
        shutil.copy2(initial_ubi_file, tmp_ubi_file)

        with open(tmp_par_file, "w") as file:
            geometry_parameters = load_par_file(geo_par_file)
            for key, value in geometry_parameters.items():
                file.write(f"{key} {value}\n")

            lattice_parameters = load_par_file(lattice_par_file)
            for key, value in lattice_parameters.items():
                file.write(f"{key} {value}\n")

        yield tmp_ubi_file, tmp_par_file
    finally:
        os.remove(tmp_ubi_file)
        os.remove(tmp_par_file)



from ewokscore import TaskWithProgress

from ImageD11 import frelon_peaksearch
from ImageD11 import columnfile as PeakColumnFile

from ..models import (
    SegmenterConfig,
    SegmenterFolderConfig,
    SegmenterCorrectionFiles,
)
from ..utils import TqdmProgressCallback, extract_sample_info
from pathlib import Path
import h5py


class SegmentScan(
    TaskWithProgress,
    input_names=["folder_config", "segmenter_algo_params", "correction_files"],
    output_names=["segmented_3d_peaks_file", "sample_folder_info"],
):
    """
    This task segments an entire scan folder,
    merges the peaks, and produces a 3D column file.
    The resulting 3D column peak file is saved.

    Outputs:

    - `segmented_3d_peaks_file`: A segmented 3d peaks for the given scan folder is saved in this path.
    - `sample_folder_info`: A Config information about raw scan sample
    """

    def run(self):

        seg_folder_config = SegmenterFolderConfig(**self.inputs.folder_config)

        detector = seg_folder_config.detector
        omega_motor = seg_folder_config.omega_motor
        scan_folder = seg_folder_config.scan_folder

        _, sample_name, dset_name, scan_number = extract_sample_info(
            path_str=scan_folder
        )

        masterfile_path = Path(scan_folder).parent / f"{sample_name}_{dset_name}.h5"
        if not masterfile_path.exists():
            raise FileNotFoundError(
                """ 'scan_folder' field in 'seg_folder_config' does not have master file
                        in its parent folder. From its parent folder, we will have its masterfile path.
                    This masterfile path required to compute Omega (rotation) values.
                """
            )
        analyse_folder = seg_folder_config.analyse_folder
        output_folder = (
            Path(analyse_folder) / sample_name / f"{sample_name}_{dset_name}"
        )
        output_folder.mkdir(parents=True, exist_ok=True)

        with h5py.File(str(masterfile_path), "r") as hin:
            omega_angles = hin[str(scan_number) + ".1"]["measurement"].get(
                omega_motor, None
            )
            omega_array = omega_angles[()]

        segmenter_cfg = SegmenterConfig(**self.inputs.segmenter_algo_params)
        correction_files_config = SegmenterCorrectionFiles(
            **self.inputs.correction_files
        )
        segmenter_settings = {
            "bgfile": correction_files_config.bg_file,
            "maskfile": correction_files_config.mask_file,
            "darkfile": correction_files_config.dark_file,
            "flatfile": correction_files_config.flat_file,
            "threshold": segmenter_cfg.threshold,
            "smoothsigma": segmenter_cfg.smooth_sigma,
            "bgc": segmenter_cfg.bgc,
            "minpx": segmenter_cfg.min_px,
            "m_offset_thresh": segmenter_cfg.offset_threshold,
            "m_ratio_thresh": segmenter_cfg.ratio_threshold,
        }

        all_frames_2d_peaks_list = frelon_peaksearch.segment_master_file(
            str(masterfile_path),
            str(scan_number) + ".1" + "/measurement/" + detector,
            omega_array,
            segmenter_settings,
            tqdm_class=TqdmProgressCallback,
            TaskInstance=self,
        )
        peaks_2d_dict, num_peaks = frelon_peaksearch.peaks_list_to_dict(
            all_frames_2d_peaks_list
        )
        # 3d merge from 2d peaks dict
        peak_3d_dict = frelon_peaksearch.do3dmerge(
            peaks_2d_dict, num_peaks, omega_array
        )
        segmented_columnfile_3d = PeakColumnFile.colfile_from_dict(peak_3d_dict)
        segmented_3d_columnfile_h5 = f"{analyse_folder}/{sample_name}/{sample_name}_{dset_name}/segmented_3d_peaks.h5"
        PeakColumnFile.colfile_to_hdf(
            segmented_columnfile_3d, segmented_3d_columnfile_h5
        )
        self.outputs.segmented_3d_peaks_file = segmented_3d_columnfile_h5
        self.outputs.sample_folder_info = {
            "omega_motor": omega_motor,
            "scan_folder": scan_folder,
        }



from ewokscore import Task
from pathlib import Path

from ImageD11 import columnfile as PeakColumnFile

from ..utils import correct_column_file


class DetectorSpatialCorrection(
    Task,
    input_names=["segmented_3d_peaks_file", "correction_files"],
    output_names=["detector_spatial_corrected_3d_peaks_file"],
):
    """
    Does the detector spatial correction on the segmented 3d peaks and saves the corrected 3D column peak file

    Inputs:

    - `segmented_3d_peaks_file`: a 3d peaks columnfile saved as hdf5 file
    - `correction_files`: two corrections are possible:
        - Spline correction: `correction_files` should be a string containing the path to the spline file
        - e2dx,e2dy correction: `correction_files` should be a tuple of 2 strings, the first one being the path to e2dx file, the second the path to the e2dy file
        - any other type will be treated as invalid input

    Outputs:
    - `detector_spatial_corrected_3d_peaks_file`: file where the resulting detector detector spatial corrected 3d peaks are saved
    """

    def run(self):
        segmented_3d_columnfile_h5 = Path(self.inputs.segmented_3d_peaks_file)
        detector_spatial_correction_file_lists = self.inputs.correction_files

        if not Path(segmented_3d_columnfile_h5).exists():
            raise FileNotFoundError(
                f""" Provided segmented 3d columnfile path: {segmented_3d_columnfile_h5} does not exist.
                """
            )

        columnfile_3d = correct_column_file(
            segmented_3d_columnfile_h5, detector_spatial_correction_file_lists
        )
        parent_folder_segmented_3d_columnfile = Path(segmented_3d_columnfile_h5).parent
        detector_spatial_corrected_3d_peaks_file = (
            parent_folder_segmented_3d_columnfile
            / "detector_spatial_corrected_3d_peaks.h5"
        )
        PeakColumnFile.colfile_to_hdf(
            columnfile_3d, str(detector_spatial_corrected_3d_peaks_file)
        )

        self.outputs.detector_spatial_corrected_3d_peaks_file = str(
            detector_spatial_corrected_3d_peaks_file
        )



from ewoks3dxrd.tasks.make_grain_map import MakeGrainMap

from .conftest import assert_grain_map_results


def test_map_grains():
    inputs = {
        "hkl_tols": (0.05, 0.025, 0.01),
        "minpks": 120,
        "lattice_name": "Fe",
        "folder_file_config": {
            "omega_motor": "diffrz",
            "scan_folder": "/data/projects/id03_3dxrd/expt/RAW_DATA/FeAu_0p5_tR/FeAu_0p5_tR_ff1/scan0001",
        },
        "flt_3dpeaks_file_for_fine_refine": "/data/projects/id03_3dxrd/ewoks_test_data/grain_map/intensity_frac_0p9837_Fe_1p0100_filtered_3d_peaks.flt",
        "indexed_grain_ubi_file": "/data/projects/id03_3dxrd/ewoks_test_data/grain_map/Fe_grains.ubi",
        "flt_3dpeaks_file": "/data/projects/id03_3dxrd/ewoks_test_data/grain_map/intensity_frac_0p9837_Fe_all_filtered_3d_peaks.flt",
    }

    task = MakeGrainMap(inputs=inputs)
    task.execute()

    assert_grain_map_results(task.outputs)



import sys

if sys.version_info >= (3, 9):
    from importlib.resources import files as resource_files
else:
    from importlib_resources import files as resource_files

from pathlib import Path

import numpy as np
from ewokscore import execute_graph
from ImageD11 import columnfile as PeakColumnFile

from orangecontrib.ewoks3dxrd import tutorials

from .conftest import assert_indexing_results, assert_grain_map_results


def test_segmentation_3dpeaks_workflow(inp_config):
    # if you change this config setting or default inp_config,
    # assertion statement assert len(found_peaks[0]) == 76 and
    # and len(peaks_position[0]) == 31682 and
    # assert cf.nrows == 7662 will fail

    # if you change this config setting,
    # assertion statement assert len(found_peaks[0]) == 76 will fail
    segmenter_config = {
        "threshold": 70,
        "smooth_sigma": 1.0,
        "bgc": 0.9,
        "min_px": 3,
        "offset_threshold": 100,
        "ratio_threshold": 150,
    }

    file_cor_config = {
        "bg_file": inp_config.get("bg_file"),
        "mask_file": inp_config.get("mask_file"),
        "flat_file": None,
        "dark_file": None,
    }

    scan_folder = Path(inp_config.get("scan_folder"))
    sample_folder_config = {
        "detector": "frelon3",
        "omega_motor": "diffrz",
        "scan_folder": str(scan_folder),
        "analyse_folder": inp_config.get("analyse_folder"),
    }

    inputs = [
        {
            "name": "folder_config",
            "value": sample_folder_config,
            "id": "segScan",
        },
        {
            "name": "segmenter_algo_params",
            "value": segmenter_config,
            "id": "segScan",
        },
        {
            "name": "correction_files",
            "value": file_cor_config,
            "id": "segScan",
        },
        {
            "name": "correction_files",
            "value": inp_config["spline_file"],
            "id": "DetCor",
        },
        {
            "name": "geometry_par_file",
            "value": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/geometry_tdxrd.par",
            "id": "GeoUp",
        },
        {
            "name": "lattice_par_file",
            "value": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/Fe.par",
            "id": "LatFilt",
        },
        {
            "name": "lattice_par_file",
            "value": "/data/projects/id03_3dxrd/expt/PROCESSED_DATA/Fe.par",
            "id": "AllLatFilt",
        },
        {
            "name": "reciprocal_dist_max",
            "value": 1.01,
            "id": "LatFilt",
        },
        {
            "name": "reciprocal_dist_tol",
            "value": 0.01,
            "id": "LatFilt",
        },
        {
            "name": "reciprocal_dist_tol",
            "value": 0.01,
            "id": "AllLatFilt",
        },
        {
            "name": "intensity_frac",
            "value": 0.9837,
            "id": "IntFilt",
        },
        {
            "name": "intensity_frac",
            "value": 0.9837,
            "id": "AllIntFilt",
        },
        {
            "name": "reciprocal_dist_tol",
            "value": 0.05,
            "id": "Indexer",
        },
        {
            "name": "gen_rings_from_idx",
            "value": (0, 1),
            "id": "Indexer",
        },
        {
            "name": "score_rings_from_idx",
            "value": (0, 1, 2, 3),
            "id": "Indexer",
        },
        {
            "name": "hkl_tols",
            "value": (0.01, 0.02, 0.03, 0.04),
            "id": "Indexer",
        },
        {
            "name": "min_pks_frac",
            "value": (0.9, 0.75),
            "id": "Indexer",
        },
        {
            "name": "cosine_tol",
            "value": np.cos(np.radians(90 - 0.25)),
            "id": "Indexer",
        },
        {
            "name": "max_grains",
            "value": 1000,
            "id": "Indexer",
        },
        {
            "name": "lattice_name",
            "value": "Fe",
            "id": "Indexer",
        },
        {
            "name": "hkl_tols",
            "value": (0.05, 0.025, 0.01),
            "id": "GrainMap",
        },
        {
            "name": "minpks",
            "value": 120,
            "id": "GrainMap",
        },
        {
            "name": "lattice_name",
            "value": "Fe",
            "id": "GrainMap",
        },
    ]
    workflow_file = resource_files(tutorials).joinpath("3dxrd_workflow.json")

    outputs = execute_graph(
        str(workflow_file), inputs=inputs, outputs=[{"all": True}], merge_outputs=False
    )
    assert outputs is not None
    detector_spatial_file_path = outputs["DetCor"][
        "detector_spatial_corrected_3d_peaks_file"
    ]
    assert Path(detector_spatial_file_path).exists()
    detector_spatial_corrected_columnfile = PeakColumnFile.columnfile(
        filename=detector_spatial_file_path
    )
    peak_x_position = detector_spatial_corrected_columnfile.f_raw
    assert len(peak_x_position) == 31682
    filtered_h5_col_file = outputs["LatFilt"]["filtered_h5_col_file"]
    filtered_flt_col_file = outputs["LatFilt"]["filtered_flt_col_file"]
    lattice_par_file = outputs["LatFilt"]["copied_lattice_par_file"]

    assert Path(filtered_h5_col_file).exists()
    assert Path(filtered_flt_col_file).exists()
    assert Path(lattice_par_file).exists()

    int_flt_h5_file = outputs["IntFilt"]["Ifilt_h5_col_file"]
    int_flt_flt_file = outputs["IntFilt"]["Ifilt_flt_col_file"]
    assert Path(int_flt_h5_file).exists()
    assert Path(int_flt_flt_file).exists()
    cf = PeakColumnFile.columnfile(filename=int_flt_h5_file)
    assert cf.nrows == 7662

    assert_indexing_results(outputs["Indexer"])

    assert_grain_map_results(outputs["GrainMap"])



import sysconfig

NAME = "Ewoks 3DXRD"

DESCRIPTION = "Ewoks 3DXRD workflows"

LONG_DESCRIPTION = "Ewoks 3DXRD workflows"

ICON = "icons/3dxrd_logo.png"

BACKGROUND = "light-blue"

WIDGET_HELP_PATH = (
    # Development documentation (make htmlhelp in ./doc)
    ("{DEVELOP_ROOT}/doc/_build/htmlhelp/index.html", None),
    # Documentation included in wheel
    ("{}/help/ewoks3dxrd/index.html".format(sysconfig.get_path("data")), None),
    # Online documentation url
    ("https://ewoks3dxrd.readthedocs.io", ""),
)



from __future__ import annotations

from pathlib import Path

from ewokscore import Task
from ImageD11 import grain as grainMod

from ..models import SampleConfig
from ..utils import (
    extract_sample_info,
    get_omega_slope,
    refine_grains,
    tmp_processing_files,
)


class MakeGrainMap(
    Task,
    input_names=[
        "folder_file_config",
        "indexed_grain_ubi_file",
        "flt_3dpeaks_file",
        "hkl_tols",
        "minpks",
        "lattice_name",
    ],
    optional_input_names=[
        "flt_3dpeaks_file_for_fine_refine",
        "intensity_two_theta_range",
        "symmetry",
    ],
    output_names=["ascii_grain_map_file"],
):
    """
    Does an iterative refinement based on `hkl_tols` followed by a fine refinement on the indexed grains

    Inputs:

    - `folder_file_config`: Output from Init Folder File Config Task as input here.
    - `indexed_grain_ubi_file`: indexed Grains ascii file path.
    - `flt_3dpeaks_file`: Filtered peaks column file path that were used for the indexing
    - `hkl_tols`: Decreasing sequence of hkl tolerances. Will be used for iterative refinement (one after the other).
    - `minpks`: To filter grains that are not associated with at least #minpks peaks after iterative refinement.
    - `lattice_name`: Use lattice parameter value from lattice par file referred by 'lattice_name'.par.

    Optional Inputs:

    - `flt_3dpeaks_file_for_fine_refine`: Peaks used to refine the grains finely at the end of the iterative refinement. Default: same file as `flt_3dpeaks_file`.
    - `intensity_two_theta_range`: tuple of two floats, giving two theta min and max when refining. Default: (0., 180.).
    - `symmetry` (str): Lattice symmetry used to further refine grains. Default: `cubic`.

    Outputs:
    - `ascii_grain_map_file`: file where the refined grains are saved
    """

    def run(self):
        ubi_grain_file = Path(self.inputs.indexed_grain_ubi_file)
        if not ubi_grain_file.exists():
            raise FileNotFoundError(
                f"Provided Indexed Ubi grain file: {ubi_grain_file} not found."
            )
        flt_pks_file = Path(self.inputs.flt_3dpeaks_file)
        if not flt_pks_file.exists():
            raise FileNotFoundError(
                f"Provided filtered Peaks File: {flt_pks_file} not found."
            )

        fine_flt_pks_file = self.get_input_value(
            "flt_3dpeaks_file_for_fine_refine", flt_pks_file
        )
        if not isinstance(fine_flt_pks_file, Path):
            fine_flt_pks_file = Path(fine_flt_pks_file)

        if not fine_flt_pks_file.exists():
            raise FileNotFoundError(
                f"Provided filtered Peaks File: {fine_flt_pks_file} not found."
            )

        cfg = SampleConfig(**self.inputs.folder_file_config)
        omega_motor = cfg.omega_motor
        scan_folder = cfg.scan_folder
        _, sample_name, dset_name, scan_number = extract_sample_info(
            path_str=scan_folder
        )
        master_file = Path(scan_folder).parent / f"{sample_name}_{dset_name}.h5"
        if not master_file.exists():
            raise FileNotFoundError(
                f"""Could not find HDF5 master file at {master_file}.
                    Check `scan_folder`, `sample_name` and `dset_name` in the `folder_file_config` input.
                """
            )

        error_message = []
        minpks = self.inputs.minpks
        if minpks <= 0:
            error_message.append("Input: 'minpks' should be a positive number")
        lattice_name = self.inputs.lattice_name
        lattice_par_file = ubi_grain_file.parent / "par_folder" / f"{lattice_name}.par"
        if not lattice_par_file.exists():
            error_message.append(
                f"Expected a lattice par file at {lattice_par_file}."
                "Perhaps the lattice_name is wrong?"
                "Or Lattice based peak filtering was not done?"
            )
        geo_par_file = ubi_grain_file.parent / "par_folder" / "geometry_tdxrd.par"
        if not geo_par_file.exists():
            error_message.append(
                f"Expected a geometry_tdxrd file at {geo_par_file}."
                "Perhaps geometry transformation was not done?"
            )
        if error_message:
            raise ValueError("\n".join(error_message))

        omega_slope = get_omega_slope(
            filepath=master_file, scan_number=scan_number, omega_motor=omega_motor
        )

        with tmp_processing_files(
            initial_ubi_file=ubi_grain_file,
            geo_par_file=geo_par_file,
            lattice_par_file=lattice_par_file,
        ) as (ubi_file, par_file):
            hkl_tols = self.inputs.hkl_tols
            intensity_tth_range = self.get_input_value(
                "intensity_two_theta_range", (0.0, 180.0)
            )
            symmetry = self.get_input_value("symmetry", "cubic")
            for tol in hkl_tols:
                iterative_refined_grains = refine_grains(
                    tolerance=tol,
                    intensity_tth_range=intensity_tth_range,
                    omega_slope=omega_slope,
                    parameter_file=par_file,
                    filtered_peaks_file=flt_pks_file,
                    ubi_file=ubi_file,
                    symmetry=symmetry,
                )
                iterative_refined_grains.savegrains(ubi_file, sort_npks=True)

            refined_grains = grainMod.read_grain_file(ubi_file)
            grains_filtered = [
                grain for grain in refined_grains if float(grain.npks) > minpks
            ]
            grainMod.write_grain_file(filename=ubi_file, list_of_grains=grains_filtered)

            # fine refinement
            fine_refined_grains = refine_grains(
                tolerance=hkl_tols[-1],
                intensity_tth_range=intensity_tth_range,
                omega_slope=omega_slope,
                parameter_file=par_file,
                filtered_peaks_file=fine_flt_pks_file,
                ubi_file=ubi_file,
                symmetry=symmetry,
            )

        output_grain_map_file = str(
            ubi_grain_file.parent / f"refined_grains_map_by_{lattice_name}.map"
        )
        fine_refined_grains.savegrains(output_grain_map_file, sort_npks=True)
        self.outputs.ascii_grain_map_file = output_grain_map_file



from __future__ import annotations

from pathlib import Path

from ewokscore import Task
from ImageD11 import columnfile as PeakColumnFile
from ImageD11 import grain, indexing
from ImageD11.unitcell import unitcell

from ..utils import (
    read_lattice_cell_data,
    read_wavelength,
    save_grains_to_nexus,
)


class IndexGrains(
    Task,
    input_names=[
        "flt_peaks_3d_file",
        "gen_rings_from_idx",
        "score_rings_from_idx",
        "lattice_name",
    ],
    optional_input_names=[
        "max_grains",
        "reciprocal_dist_tol",
        "hkl_tols",
        "min_pks_frac",
        "cosine_tol",
    ],
    output_names=["grain_ubi_file", "grain_nexus_file"],
):
    """
    From 3D peaks, finds grains' UBI  matrices and stores them in both ASCII format and NeXus (.h5) format.

    Inputs:

    - `flt_peaks_3d_file` (str): File containing lattice and intensity filtered 3d merged peaks.
    - `gen_rings_from_idx` (Tuple): Indices of rings used for generating UBI. Two are usually enough, three in some rare cases.
    - `score_rings_from_idx` (Tuple): Indices of the rings used for scoring. These must contain the indices used for indexing.
    - `lattice_name` (str): lattice name indicating {lattice_name}.par file
            a parameter file inside the par_folder

    Optional Inputs:

    - `max_grains` (int): To limit the maximum number of grains (UBI matrices).
    - `reciprocal_dist_tol` (float): reciprocal distance tolerance value.
    - `hkl_tols` (Tuple): hkl tolerance, (hkl are integers, while doing convergence, had to do discretization on processed values)
    - `min_pks_frac` (Tuple): min peaks fraction to iterate over
    - `cosine_tol` (float): a tolerance value used in the Indexer convergence scheme
        for finding pairs of peaks to make an orientation

    Outputs:

    - `grain_ubi_file` (str): Path to the ASCII file that stores the generated UBI
    - `grain_nexus_file` (str): Path to the NeXus file that stores the computed grains along with lattice parameter,
            geometry_tdxrd information etc.
    """

    def run(self):
        ph_flt_h5_file = Path(self.inputs.flt_peaks_3d_file)
        gen_rings_idx = self.inputs.gen_rings_from_idx
        score_rings_idx = self.inputs.score_rings_from_idx
        lattice_name = self.inputs.lattice_name

        indexing_params = {}
        if self.get_input_value("max_grains", None):
            indexing_params["max_grains"] = self.inputs.max_grains
        if self.get_input_value("reciprocal_dist_tol", None):
            indexing_params["dstol"] = self.inputs.reciprocal_dist_tol
        if self.get_input_value("hkl_tols", None):
            indexing_params["hkl_tols"] = self.inputs.hkl_tols
        if self.get_input_value("min_pks_frac", None):
            indexing_params["fracs"] = self.inputs.min_pks_frac
        if self.get_input_value("cosine_tol", None):
            indexing_params["cosine_tol"] = self.inputs.cosine_tol

        error_message = []

        if not ph_flt_h5_file.exists():
            error_message.append(
                f"Filtered 3d peaks file '{ph_flt_h5_file}' not found."
            )
        if len(gen_rings_idx) < 2:
            error_message.append(
                f"UBI needs at least two ring indices in `gen_rings_idx`. Got {gen_rings_idx}"
            )

        work_folder = ph_flt_h5_file.parent
        lattice_par_file = work_folder / f"par_folder/{lattice_name}.par"
        if not lattice_par_file.exists():
            error_message.append(
                f"Expected a lattice par file at {lattice_par_file}. Perhaps the lattice_name is wrong?"
                "in the parent folder of provided lattice filtered 3d peaks file: "
                f"{ph_flt_h5_file}"
            )
        if error_message:
            raise ValueError("\n".join(error_message))

        filtered_cf = PeakColumnFile.columnfile(filename=str(ph_flt_h5_file))
        unit_cell_parameters = read_lattice_cell_data(
            lattice_data_file_path=lattice_par_file
        )
        unit_cell = unitcell(
            lattice_parameters=unit_cell_parameters.lattice_parameters,
            symmetry=unit_cell_parameters.space_group,
        )
        unit_cell.makerings(limit=filtered_cf.ds.max())
        geometry_par_file = work_folder / "par_folder" / "geometry_tdxrd.par"
        if not geometry_par_file.exists():
            raise FileNotFoundError(f"Expected a geometry file at {geometry_par_file}")
        wavelength = read_wavelength(geometry_par_file)
        grains, _ = indexing.do_index(
            cf=filtered_cf,
            forgen=gen_rings_idx,
            foridx=score_rings_idx,
            unitcell=unit_cell,
            wavelength=wavelength,
            **indexing_params,
        )

        # have to save the grains, and ubi matrices.
        grain_ubi_file_path = str(work_folder / f"{lattice_name}_grains.ubi")
        grain.write_grain_file(grain_ubi_file_path, grains)
        grain_nexus_file = str(work_folder / f"{lattice_name}_grains.h5")
        save_grains_to_nexus(
            nexus_filename=grain_nexus_file,
            geometry_par_path=geometry_par_file,
            lattice_par_file=lattice_par_file,
            lattice_name=lattice_name,
            grains=grains,
        )
        self.outputs.grain_ubi_file = grain_ubi_file_path
        self.outputs.grain_nexus_file = grain_nexus_file



from ewokscore import Task
from pathlib import Path
import shutil

from ImageD11 import columnfile as PeakColumnFile
from ..utils import read_lattice_cell_data
from ImageD11.unitcell import unitcell
from ImageD11.peakselect import filter_peaks_by_phase


class FilterByLattice(
    Task,
    input_names=[
        "geometry_trans_3d_peaks_file",
        "lattice_par_file",
        "reciprocal_dist_tol",
    ],
    optional_input_names=["reciprocal_dist_max"],
    output_names=[
        "filtered_h5_col_file",
        "filtered_flt_col_file",
        "copied_lattice_par_file",
    ],
):
    """
    Performs Lattice/Phase-based filtering on a geometry-transformed 3D peaks file.

    This process applies filtering based on 'reciprocal distance' and 'lattice' rings ds criteria
    to extract relevant peaks.

    ### Steps:
    1. **Initial Filtering:**
        - Copies the input geometry-transformed 3D peaks file.
        - Reads the `ds` column and removes rows where `ds` exceeds the specified `reciprocal_dist_max` value.

    2. **Lattice-Based Filtering:**
        - Computes ideal lattice ring `ds` values (reciprocal distances from the origin).
        - Further filters peaks based on these values, using the tolerance defined by `reciprocal_dist_tol`.

    3. **File Storage:**
        - Saves the lattice-filtered 3D peaks file as
            `{Lattice_name}_{reciprocal_dist_max_tag}_filtered_3d_peaks.h5,flt`
            in the parent directory of the input file ('geometry_trans_3d_peaks_file').

    Additionally, if the specified `lattice_par_file` is not present in the sample analysis path,
    it is copied to `"par_folder/{lattice_par_file}"`.

    ### Inputs
    - `geometry_trans_3d_peaks_file` (str): Path to the geometry-transformed `.h5` 3D peaks file.
    - `lattice_par_file` (str): Path to the `.par` file containing lattice parameters and space group information.
    - `reciprocal_dist_max` (float): Maximum reciprocal distance for filtering (an Optional Value)
        If it is not provided, then maximum value in the `ds` column from input file will be used.
    - `reciprocal_dist_tol` (float): Tolerance for peak inclusion near lattice rings.

    ### Outputs
    - `filtered_h5_col_file` (str): Path to the lattice-filtered `.h5` file, named as
    `{lattice_name}_{reciprocal_dist_max_tag}_filtered_3d_peaks.h5`.
    The `reciprocal_dist_max_tag` is either `'all'` or formatted as `f"{dstar_max:.4f}".replace(".", "p")`.
    - `filtered_flt_col_file` (str): Path to the same filtered file in ASCII format (`.flt`).
    - `copied_lattice_par_file` (str): Path to the copied lattice parameter file stored within the analysis folder.
    """

    def run(self):
        geometry_trans_3d_peaks_file = Path(self.inputs.geometry_trans_3d_peaks_file)
        lattice_par_file = Path(self.inputs.lattice_par_file)
        dstar_tol = self.inputs.reciprocal_dist_tol
        error_message = []

        if not geometry_trans_3d_peaks_file.exists():
            error_message.append(
                f"Geometry file '{geometry_trans_3d_peaks_file}' not found."
            )
        if not lattice_par_file.exists():
            error_message.append(
                f"Provided Lattice file '{lattice_par_file}' does not exist."
            )
        if lattice_par_file.suffix != ".par":
            error_message.append(
                f"Provided Lattice file '{lattice_par_file}' is not a .par file"
            )
        if error_message:
            raise ValueError("\n".join(error_message))

        lattice_file_in_par_folder = (
            geometry_trans_3d_peaks_file.parent / "par_folder" / lattice_par_file.name
        )

        lattice_file_in_par_folder.parent.mkdir(exist_ok=True)
        shutil.copy2(lattice_par_file, lattice_file_in_par_folder)

        g_trans_peaks_3d = PeakColumnFile.columnfile(
            filename=str(geometry_trans_3d_peaks_file)
        )
        unit_cell_parameters = read_lattice_cell_data(lattice_par_file)

        lattice_name = lattice_par_file.stem
        reciprocal_dist_max: float | None = self.get_input_value(
            "reciprocal_dist_max", None
        )
        if reciprocal_dist_max:
            dstar_max = reciprocal_dist_max
            dstar_as_str = f"{dstar_max:.4f}".replace(".", "p")
            flt_file_name = f"{lattice_name}_{dstar_as_str}_filtered_3d_peaks.flt"

        else:
            dstar_max = g_trans_peaks_3d.ds.max()
            flt_file_name = f"{lattice_name}_all_filtered_3d_peaks.flt"

        cf = g_trans_peaks_3d.copyrows(g_trans_peaks_3d.ds <= dstar_max)
        ucell = unitcell(
            lattice_parameters=unit_cell_parameters.lattice_parameters,
            symmetry=unit_cell_parameters.space_group,
        )
        ucell.makerings(limit=g_trans_peaks_3d.ds.max())

        filtered_peaks_cf = filter_peaks_by_phase(
            cf=cf, dstol=dstar_tol, dsmax=dstar_max, cell=ucell
        )

        flt_file_path = geometry_trans_3d_peaks_file.parent / flt_file_name
        filtered_peaks_cf.writefile(str(flt_file_path))
        h5_file_path = flt_file_path.with_suffix(".h5")
        PeakColumnFile.colfile_to_hdf(filtered_peaks_cf, str(h5_file_path))

        self.outputs.filtered_h5_col_file = str(h5_file_path)
        self.outputs.filtered_flt_col_file = str(flt_file_path)
        self.outputs.copied_lattice_par_file = str(lattice_file_in_par_folder)



from __future__ import annotations

from pathlib import Path
from typing import Any, List, Tuple

import h5py
import numpy as np
from ImageD11 import grain as grainmod
from ImageD11.grain import grain as Grain


def create_parameters_group(
    parent_group: h5py.Group,
    config_settings: dict[str, Any],
    group_name: str = "parameters",
):
    """
    Helper function to place the config_settings in 'param_name' group
    in the given `parent_group`
    """
    parameters_group = parent_group.create_group(group_name)
    parameters_group.attrs["NX_class"] = "NXcollection"
    for key, value in config_settings.items():
        parameters_group.create_dataset(
            key, data=value if value is not None else "None"
        )


def save_nexus_segmentation(
    entry_name: str,
    filename: str | Path,
    seg_3d_data: dict[str, np.ndarray],
    config_set: dict[str, dict[str, Any]],
):
    """
    Save a 3D segmentation dataset into an HDF5 (NeXus) file.
    Inputs:
    - `entry_name`: Name of the parent entry where the segmenter process (group) will be saved.
    - `filename`: Path to the HDF5 file where data will be saved.
    - `seg_3d_data`: peaks column data sets.
    - `config_set`: Configuration parameters structured as nested dictionaries.
    """
    with h5py.File(filename, "w") as f:
        entry = f.create_group(entry_name)
        entry.attrs["NX_class"] = "NXentry"

        segmentation_group = entry.create_group("segmented_3d_peaks")
        segmentation_group.attrs["NX_class"] = "NXprocess"

        peaks_group = segmentation_group.create_group("peaks")
        peaks_group.attrs["NX_class"] = "NXdata"

        for key, value in seg_3d_data.items():
            peaks_group.create_dataset(key, data=value)

        parameters_group = segmentation_group.create_group("parameters")
        parameters_group.attrs["NX_class"] = "NXcollection"

        for param_name, param_values in config_set.items():
            create_parameters_group(
                parent_group=parameters_group,
                config_settings=param_values,
                group_name=param_name,
            )


def create_peaks_column_group(
    filename: str | Path,
    entry_name: str,
    process_name: str,
    peaks_data: dict[str, np.ndarray],
    config_settings: dict[str, Any],
):
    """
    Create a NXprocess group to save peaks data and parameters
    Inputs:
        `filename`: file path to existing .h5 file
        `entry_name`: entry point in the .h5 file
        `data_group`: new group name inside the sample_dataset
        `peaks_data`: peaks column data sets
        `config_settings`: dict of configuration settings.
    """
    with h5py.File(filename, "a") as f:
        entry = f[entry_name]
        grp = entry.create_group(process_name)
        grp.attrs["NX_class"] = "NXprocess"
        peaks_group = grp.create_group("peaks")
        for key, value in peaks_data.items():
            peaks_group.create_dataset(key, data=value)

        create_parameters_group(parent_group=grp, config_settings=config_settings)


def read_peaks_attributes(
    filename: str | Path, entry_name: str, process_name: str
) -> dict[str, np.ndarray]:
    """
    Extract peaks column data stored in {entry_name}/{process_name}/peaks
    Inputs:
        filename: file path to .h5 file
        entry_name: entry point inside .h5 file
        process_name: group name inside the entry point
    """
    with h5py.File(filename, "r") as f:
        peaks_group = f[f"{entry_name}/{process_name}/peaks"]
        return {name: dataset[()] for name, dataset in peaks_group.items()}


def create_nexus_ubi(
    grains: List[Grain],
    entry_name: str,
    grain_file: str | Path,
    grain_settings: dict[str, Any],
    peaks_path: Tuple[str | Path, str],
):
    """
    Create nexus file with entry_name place the grains (list of UBIs)
    and also link the peaks used to generate these UBIS as external link
    """
    with h5py.File(grain_file, "w") as gf:
        entry = gf.create_group(entry_name)
        entry.attrs["NX_class"] = "NXentry"
        grain_group = entry.create_group("indexed_grains")
        grain_group.attrs["NX_class"] = "NXprocess"
        grain_group["peaks"] = h5py.ExternalLink(peaks_path[0], peaks_path[1])
        ubi_group = grain_group.create_group("UBI")
        ubi_group.attrs["NX_class"] = "NXdata"
        ubi_matrices = ubi_group.create_dataset(
            "UBI", shape=(len(grains), 3, 3), dtype=grains[0].ubi.dtype
        )
        for i, grain in enumerate(grains):
            ubi_matrices[i] = grain.ubi
        create_parameters_group(
            parent_group=grain_group, config_settings=grain_settings
        )


def create_nexus_grains(
    grains: List[Grain],
    entry_name: str,
    grain_file: str | Path,
    grain_settings: dict[str, Any],
    grain_group_name: str,
):
    num_grains = len(grains)
    with h5py.File(grain_file, "a") as gf:
        entry = gf[entry_name]
        grp = entry.create_group(grain_group_name)
        grp.attrs["NX_class"] = "NXprocess"
        grains_gr = grp.create_group("grains")
        grains_gr.attrs["NX_class"] = "NXdata"

        ubi_matrices = grains_gr.create_dataset(
            "UBI", shape=(num_grains, 3, 3), dtype=grains[0].ubi.dtype
        )
        translations = grains_gr.create_dataset(
            "translation", shape=(num_grains, 3), dtype=grains[0].ubi.dtype
        )
        npks = grains_gr.create_dataset("npks", shape=(num_grains,), dtype=np.int64)
        nuniq = grains_gr.create_dataset("nuniq", shape=(num_grains,), dtype=np.int64)
        for i, grain in enumerate(grains):
            ubi_matrices[i] = grain.ubi
            translations[i] = grain.translation
            npks[i] = grain.npks
            nuniq[i] = grain.nuniq

        create_parameters_group(parent_group=grp, config_settings=grain_settings)


def save_indexed_grains_as_ascii(
    grain_file_h5: str | Path,
    entry_name: str,
    process_group_name: str,
    grain_file_ascii: str | Path,
):
    """
    Function to extract grains that generated by Indexing function,
    which is simpler, i.e it only has UBI matrices.
    """
    grains_list = []
    with h5py.File(grain_file_h5, "r") as gf:
        entry = gf[f"{entry_name}/{process_group_name}"]
        ubi_grp = entry["UBI"]
        for ubi_matrix in ubi_grp["UBI"]:
            grains_list.append(Grain(ubi=ubi_matrix))

    grainmod.write_grain_file(grain_file_ascii, grains_list)


def read_grains(
    grain_file_h5: str | Path, entry_name: str, process_group_name: str
) -> List[Grain]:

    grains_list = []
    with h5py.File(grain_file_h5, "r") as gf:
        entry = gf[f"{entry_name}/{process_group_name}/grains"]
        for i in range(entry["translation"][()].shape[0]):
            gr = Grain.grain(ubi=entry["UBI"][i], translation=entry["translation"][i])
            gr.npks = entry["npks"][i]
            gr.nuniq = entry["nuniq"][i]
            grains_list.append(gr)
    return grains_list


def get_omega_array(
    filename: str | Path, entry_name: str, process_group_name: str
) -> np.ndarray:
    with h5py.File(filename, "r") as f:
        entry = f[f"{entry_name}/{process_group_name}"]
        folder_grp = entry["parameters/FolderFileSettings"]
        masterfile = folder_grp["masterfile"][()].decode()
        scan_number = folder_grp["scan_number"][()].decode()
        omegamotor = folder_grp["omegamotor"][()].decode()
        with h5py.File(masterfile, "r") as hin:
            omega_angles = hin[f"{scan_number}.1/measurement"][omegamotor]
            return omega_angles[()]


def get_lattice_parameters(
    filename: str | Path, entry_name: str, process_group_name: str
) -> Tuple[np.ndarray, int]:
    with h5py.File(filename, "r") as f:
        entry = f[f"{entry_name}/{process_group_name}"]
        par_grp = entry["parameters"]
        lattice_parameters = par_grp["lattice_parameters"][()]
        symmetry = int(par_grp["lattice_space_group"][()])
    return lattice_parameters, symmetry


def get_wavelength(
    filename: str | Path, entry_name: str, process_group_name: str
) -> float:
    with h5py.File(filename, "r") as f:
        entry = f[f"{entry_name}/{process_group_name}"]
        par_grp = entry["parameters"]
        wavelength = float(par_grp["wavelength"][()])
    return wavelength


def get_parameters(
    filename: str | Path, entry_name: str, process_group_name: str
) -> dict[str, str]:
    par_dict = {}
    with h5py.File(filename, "r") as f:
        entry = f[f"{entry_name}/{process_group_name}"]
        par_grp = entry["parameters"]
        for key in par_grp.keys():
            value = par_grp[key][()]
            par_dict[key] = value.decode() if isinstance(value, bytes) else value
    return par_dict


def save_geometry_and_lattice_par_file(
    file_path: str | Path, geom_dict: dict[str, Any], lattice_dict: dict[str, Any]
):
    with open(file_path, "w") as f:
        lattice_params = lattice_dict.get("lattice_parameters", [])
        if len(lattice_params) != 6:
            raise ValueError(
                f"Expected a list of 6 lattice params. Got {lattice_params}"
            )
        f.write(f"cell__a {lattice_params[0]}\n")
        f.write(f"cell__b {lattice_params[1]}\n")
        f.write(f"cell__c {lattice_params[2]}\n")
        f.write(f"cell_alpha {lattice_params[3]}\n")
        f.write(f"cell_beta {lattice_params[4]}\n")
        f.write(f"cell_gamma {lattice_params[5]}\n")

        if "lattice_space_group" in lattice_dict:
            f.write(
                f"cell_lattice_[P,A,B,C,I,F,R] {lattice_dict['lattice_space_group']}\n"
            )

        for key, value in geom_dict.items():
            f.write(f"{key} {value}\n")
